import os
import collections
import json
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from PIL import Image
import collections
import torch
import torchvision

from torch.utils import data
import random 

from scipy.io import loadmat
import pickle5 as pickle
from sklearn.preprocessing import MinMaxScaler

def location_cal(sensor_positions, batch_features):
    NumChannels = batch_features.shape[1]
    location_day = []

    b_max = batch_features.max(-1)
    b_min = batch_features.min(-1)
    amplitudes = b_max-b_min
    # amplitudes_multi = np.multiply(amplitudes,amplitudes)
    # amplitudes = np.multiply(amplitudes_multi,amplitudes)
    amplitudes =np.square(amplitudes)
    amplitudes = np.square(amplitudes)
    sum_square_amplitute=np.sum(amplitudes,axis=1)

    location_day=[]
    for ij in range(sensor_positions.shape[1]):
        x=np.dot(sensor_positions[:, ij] , amplitudes.T)
        x=np.divide(x, sum_square_amplitute)
        location_day.append(x)
    # y=np.dot(sensor_positions[:, 1] , amplitudes.T)
    # y=np.divide(y, sum_square_amplitute)
    #
    # # location_day = [x, y]
    # z=np.dot(sensor_positions[:, 2] , amplitudes.T)
    # z=np.divide(z, sum_square_amplitute)
    # location_day=[x,y,z]
    location_day=np.array(location_day).T
    return location_day


def location_cal_group(sensor_positions, batch_features,group_id):
    group_batch = sensor_positions[:,-1]
    location_day=np.zeros((batch_features.shape[0],3))
    for i in np.unique(group_batch):
        care_loc = np.where(group_batch==i)[0]
        look_spike_loc = np.nonzero(np.in1d(group_id, care_loc))[0]
        location_day_batch = location_cal(sensor_positions[care_loc,:], batch_features[look_spike_loc,:,:][:,care_loc,:])
        location_day[look_spike_loc,:] = location_day_batch
    return location_day



class waveformLoader(data.Dataset):
    def __init__(self, root, Keep_id , shank_channel ,
                 sensor_positions, splitgroup=False):
        with (open(root + "X_waveform.pkl", "rb")) as openfile:
            datafile = pickle.load(openfile)
        with (open(root + "Y_spike_id.pkl", "rb")) as openfile:
            GT = pickle.load(openfile)
        with (open(root + "Y_spike_id_noise.pkl", "rb")) as openfile:
            channel_id = np.array(pickle.load(openfile))

        ### datafile = datafile[:, shank_channel, :]
        for i in Keep_id:
            if i not in GT:
                keep_index = np.in1d(GT, Keep_id + [-1])
                datafile = datafile[keep_index, :, :]
                channel_id = channel_id[keep_index]
                GT = np.array(GT)[keep_index]
                break
        #
        # datafile = datafile[[ij in shank_channel for ij in channel_id], :, :]
        # GT = GT[[ij in shank_channel for ij in channel_id]]
        # channel_id = channel_id[[ij in shank_channel for ij in channel_id]]

        mask = ~np.isin(GT, Keep_id)
        GT = np.array(GT)

        GT_binary = np.zeros((GT.shape[0], 2))
        GT_binary[list(mask), 0] = 1
        GT_binary[~mask, 1] = 1

        self.GT_unique = Keep_id + [-1]
        self.GT_binary = GT_binary

        # for ind, i in enumerate(shank_channel):
        #     channel_id[channel_id == i] = ind
        self.Img_single = datafile[np.arange(datafile.shape[0]), np.array(channel_id).astype('int'), :]


        self.GT_LIST = GT

        GT_array = np.zeros((len(GT), len(self.GT_unique)))
        for idx, unique_id in enumerate(self.GT_unique):
            rmv_list = np.where(np.array(GT) == unique_id)[0]
            GT_array[rmv_list, idx] = 1

        self.GT = GT_array

        self.Img = datafile

        if splitgroup:
            pred_location = location_cal_group(sensor_positions, datafile, channel_id)
        else:
            pred_location = location_cal(sensor_positions, datafile)
        self.pred_location = pred_location
        print('pred_location',pred_location.shape)
        self.n_classes = len(set(self.GT_unique))


        # # if ch_group!=None:
        # #     test = np.array([ch_group[i] for i in channel_id])
        # #     pred_location = np.append(pred_location,test.reshape(-1,1),1)





        # i=0909
        # GT1=np.argmax(self.GT,axis=1)
        # for yichun_j in np.unique(GT1):
        #     plt.figure(figsize=(10,1))
        #     test = np.where(GT1==yichun_j)[0]
        #     plt.plot(np.arange(900),np.mean(self.Img[test,:],axis=0).flatten())
        #     for i in range(0,900,30):
        #         plt.vlines(i, -150, +150, linestyles='solid', colors='k', alpha=0.2)
        #
        #     plt.vlines(30*int(channel_id[test[0]]), -100, +100, linestyles='solid', colors='red')
        #     plt.title(yichun_j)
        #     plt.show()
        #
        # #
        # # plt.figure(figsize=(10,3))
        # # test = np.where(GT==yichun_j)[0]
        # # for yichun_i in range(5000):
        # #     plt.plot(np.arange(30), datafile[test[yichun_i],int(channel_id[test[yichun_i]]), :, ])
        # #     # break
        # # plt.show()
        #
        # # return self.Img[index, ...] ,  self.GT[index, ...], self.GT_binary[index, ...], self.Img_single[index, ...],self.pred_location[index,...]
        #
        # GT1 = np.argmax(self.GT, axis=1)
        # for yichun_j in set(GT1):
        #     plt.figure()
        #     test = np.where(GT1==yichun_j)[0]
        #     for yichun_i in range(1000):
        #         plt.plot(np.arange(30), self.Img_single[test[yichun_i],:, ])
        #         # break
        #     plt.title(yichun_j)
        #     plt.show()

        # sensor_positions
        ##### plot prediction spike
        # spike_id=np.argmax(GT_array,axis=1)
        #
        # plt.figure(figsize=(5,7))
        # color_list=np.random.rand(30,3)
        # # color_list=[[0,166,81],[236,0,140],[0,174,239],[241,90,41],[141,198,63],[102,45,145]]
        # # color_list=np.array(color_list)/255
        # for i in range(30):
        #     plt.scatter(sensor_positions[i,0],sensor_positions[i,1],c='lightgrey',
        #                 edgecolors=color_list[i],linewidths=10,
        #                 s=50)
        #
        # for i, txt in enumerate(range(30)):
        #     plt.annotate(txt, (sensor_positions[i,0], sensor_positions[i,1]))
        #
        # for i in range(30):
        #     idx = np.where(spike_id == i)
        #     plt.scatter(np.mean(self.pred_location[idx, 0]),
        #             np.mean(self.pred_location[idx, 1]), edgecolors='black',
        #             c=color_list[i], s=5)
        # # for i in range(6, 7):
        # #     idx = np.where(spike_id == i)
        # #     plt.scatter(self.pred_location[idx, 0],
        # #                 self.pred_location[idx, 1],
        # #                 c='k', s=0.0001)
        # # plt.xlim([-10,60])
        # # plt.ylim([-20, 120])
        # plt.axis('off')
        #
        # # plt.savefig('mouse6_shank1.png',dpi=300,transparent=True)
        # plt.show()



        # for ind, i in enumerate(shank_channel):
        #     channel_id[channel_id==i]=ind
        # self.GT_LIST = GT
        # self.n_classes = len(set(self.GT_unique))


    def __len__(self):
        return len(self.GT)



    def __getitem__(self, index):

        # img = self.Img[index, ...]
        # lbl = self.GT[index, ...]
        # lbl_binary = self.GT_binary[index, ...]


        return self.Img[index, ...] ,  self.GT[index, ...], self.GT_binary[index, ...], self.Img_single[index, ...],self.pred_location[index,...]


        ##### plot prediction spike
        # spike_id=np.argmax(GT_array,axis=1)
        #
        # plt.figure(figsize=(5,7))
        # color_list=[[0,166,81],[236,0,140],[0,174,239],[241,90,41],[141,198,63],[102,45,145]]
        # color_list=np.array(color_list)/255
        # for i in range(6):
        #     plt.scatter(sensor_positions[i,0],sensor_positions[i,1],c='lightgrey',
        #                 edgecolors=color_list[i],linewidths=10,
        #                 s=5000)
        #
        # for i in range(6):
        #     idx=np.where(spike_id==i)
        #     plt.scatter(self.pred_location[idx,0],
        #                 self.pred_location[idx,1],
        #                 c=color_list[i],s=0.001)
        # for i in range(6):
        #     idx = np.where(spike_id == i)
        #     plt.scatter(np.mean(self.pred_location[idx, 0]),
        #             np.mean(self.pred_location[idx, 1]), edgecolors='black',
        #             c=color_list[i], s=500)
        # # for i in range(6, 7):
        # #     idx = np.where(spike_id == i)
        # #     plt.scatter(self.pred_location[idx, 0],
        # #                 self.pred_location[idx, 1],
        # #                 c='k', s=0.001)
        # plt.xlim([-10,60])
        # plt.ylim([-20, 120])
        # plt.axis('off')
        #
        # plt.savefig('mouse6_shank1.png',dpi=300,transparent=True)
        # # plt.savefig('mouse6_shank1_noise.png',dpi=300,transparent=True)
        # plt.show()


#! /bin/env python
#
# Michael Gibson 23 April 2015


import sys, struct, os

def read_qstring(fid):
    """Read Qt style QString.  

    The first 32-bit unsigned number indicates the length of the string (in bytes).  
    If this number equals 0xFFFFFFFF, the string is null.

    Strings are stored as unicode.
    """

    length, = struct.unpack('<I', fid.read(4))
    if length == int('ffffffff', 16): return ""

    if length > (os.fstat(fid.fileno()).st_size - fid.tell() + 1) :
        print(length)
        raise Exception('Length too long.')

    # convert length from bytes to 16-bit Unicode words
    length = int(length / 2)

    data = []
    for i in range(0, length):
        c, = struct.unpack('<H', fid.read(2))
        data.append(c)

    if sys.version_info >= (3,0):
        a = ''.join([chr(c) for c in data])
    else:
        a = ''.join([unichr(c) for c in data])
    
    return a
  
if __name__ == '__main__':
    a=read_qstring(open(sys.argv[1], 'rb'))
    print(a)



#! /bin/env python
#
# Michael Gibson 27 April 2015
# Modified Adrian Foy Sep 2018

def data_to_result(header, data, data_present):
    """Moves the header and data (if present) into a common object."""
    
    result = {}
    if header['num_amplifier_channels'] > 0 and data_present:
        result['t_amplifier'] = data['t_amplifier']
    if header['num_aux_input_channels'] > 0 and data_present:
        result['t_aux_input'] = data['t_aux_input']
    if header['num_supply_voltage_channels'] > 0 and data_present:
        result['t_supply_voltage'] = data['t_supply_voltage']
    if header['num_board_adc_channels'] > 0 and data_present:
        result['t_board_adc'] = data['t_board_adc']
    if (header['num_board_dig_in_channels'] > 0 or header['num_board_dig_out_channels'] > 0) and data_present:
        result['t_dig'] = data['t_dig']
    if header['num_temp_sensor_channels'] > 0 and data_present:
        result['t_temp_sensor'] = data['t_temp_sensor']
        
    if header['num_amplifier_channels'] > 0:
        result['spike_triggers'] = header['spike_triggers']
        
    result['notes'] = header['notes']
    result['frequency_parameters'] = header['frequency_parameters']
    
    if header['version']['major'] > 1:
        result['reference_channel'] = header['reference_channel']
    
    if header['num_amplifier_channels'] > 0:
        result['amplifier_channels'] = header['amplifier_channels']
        if data_present:
            result['amplifier_data'] = data['amplifier_data']
            
    if header['num_aux_input_channels'] > 0:
        result['aux_input_channels'] = header['aux_input_channels']
        if data_present:
            result['aux_input_data'] = data['aux_input_data']
            
    if header['num_supply_voltage_channels'] > 0:
        result['supply_voltage_channels'] = header['supply_voltage_channels']
        if data_present:
            result['supply_voltage_data'] = data['supply_voltage_data']
    
    if header['num_board_adc_channels'] > 0:
        result['board_adc_channels'] = header['board_adc_channels']
        if data_present:
            result['board_adc_data'] = data['board_adc_data']
            
    if header['num_board_dig_in_channels'] > 0:
        result['board_dig_in_channels'] = header['board_dig_in_channels']
        if data_present:
            result['board_dig_in_data'] = data['board_dig_in_data']
            
    if header['num_board_dig_out_channels'] > 0:
        result['board_dig_out_channels'] = header['board_dig_out_channels']
        if data_present:
            result['board_dig_out_data'] = data['board_dig_out_data']
    
    return result


import sys, math, os, time
import os
from intanutil.read_header import read_header
from intanutil.get_bytes_per_data_block import get_bytes_per_data_block
from intanutil.read_one_data_block import read_one_data_block
from intanutil.notch_filter import notch_filter
from intanutil.data_to_result import data_to_result
import os
import spikeinterface
import spikeinterface.extractors as se
import spikeinterface.sorters as ss
import spikeinterface.widgets as sw
from spikeinterface.core.npzsortingextractor import NpzSortingExtractor
from pylab import *
ss.Kilosort3Sorter.set_kilosort3_path('/kilosort3')
import pylab
from probeinterface import Probe
import copy
import sys
sys.path.append("./autosort/util/") 
import scipy
import pickle
from collections import Counter
import numpy as np


def save_obj(objt, name):
    with(open(name+'.pkl','wb')) as f:
        pickle.dump(objt,f,pickle.HIGHEST_PROTOCOL)
        f.close()



def read_data(filename):
    """Reads Intan Technologies RHD2000 data file generated by evaluation board GUI.

    Data are returned in a dictionary, for future extensibility.
    """

    tic = time.time()
    fid = open(filename, 'rb')
    filesize = os.path.getsize(filename)

    header = read_header(fid)

    print('Found {} amplifier channel{}.'.format(header['num_amplifier_channels'],
                                                 plural(header['num_amplifier_channels'])))
    print('Found {} auxiliary input channel{}.'.format(header['num_aux_input_channels'],
                                                       plural(header['num_aux_input_channels'])))
    print('Found {} supply voltage channel{}.'.format(header['num_supply_voltage_channels'],
                                                      plural(header['num_supply_voltage_channels'])))
    print('Found {} board ADC channel{}.'.format(header['num_board_adc_channels'],
                                                 plural(header['num_board_adc_channels'])))
    print('Found {} board digital input channel{}.'.format(header['num_board_dig_in_channels'],
                                                           plural(header['num_board_dig_in_channels'])))
    print('Found {} board digital output channel{}.'.format(header['num_board_dig_out_channels'],
                                                            plural(header['num_board_dig_out_channels'])))
    print('Found {} temperature sensors channel{}.'.format(header['num_temp_sensor_channels'],
                                                           plural(header['num_temp_sensor_channels'])))
    print('')

    # Determine how many samples the data file contains.
    bytes_per_block = get_bytes_per_data_block(header)

    # How many data blocks remain in this file?
    data_present = False
    bytes_remaining = filesize - fid.tell()
    if bytes_remaining > 0:
        data_present = True

    if bytes_remaining % bytes_per_block != 0:
        raise Exception('Something is wrong with file size : should have a whole number of data blocks')

    num_data_blocks = int(bytes_remaining / bytes_per_block)

    num_amplifier_samples = header['num_samples_per_data_block'] * num_data_blocks
    num_aux_input_samples = int((header['num_samples_per_data_block'] / 4) * num_data_blocks)
    num_supply_voltage_samples = 1 * num_data_blocks
    num_board_adc_samples = header['num_samples_per_data_block'] * num_data_blocks
    num_board_dig_in_samples = header['num_samples_per_data_block'] * num_data_blocks
    num_board_dig_out_samples = header['num_samples_per_data_block'] * num_data_blocks

    record_time = num_amplifier_samples / header['sample_rate']

    if data_present:
        print('File contains {:0.3f} seconds of data.  Amplifiers were sampled at {:0.2f} kS/s.'.format(record_time,
                                                                                                        header[
                                                                                                            'sample_rate'] / 1000))
    else:
        print('Header file contains no data.  Amplifiers were sampled at {:0.2f} kS/s.'.format(
            header['sample_rate'] / 1000))

    if data_present:
        # Pre-allocate memory for data.
        print('')
        print('Allocating memory for data...')

        data = {}
        if (header['version']['major'] == 1 and header['version']['minor'] >= 2) or (header['version']['major'] > 1):
            data['t_amplifier'] = np.zeros(num_amplifier_samples, dtype=np.int)
        else:
            data['t_amplifier'] = np.zeros(num_amplifier_samples, dtype=np.uint)

        data['amplifier_data'] = np.zeros([header['num_amplifier_channels'], num_amplifier_samples], dtype=np.uint)
        data['aux_input_data'] = np.zeros([header['num_aux_input_channels'], num_aux_input_samples], dtype=np.uint)
        data['supply_voltage_data'] = np.zeros([header['num_supply_voltage_channels'], num_supply_voltage_samples],
                                               dtype=np.uint)
        data['temp_sensor_data'] = np.zeros([header['num_temp_sensor_channels'], num_supply_voltage_samples],
                                            dtype=np.uint)
        data['board_adc_data'] = np.zeros([header['num_board_adc_channels'], num_board_adc_samples], dtype=np.uint)

        # by default, this script interprets digital events (digital inputs and outputs) as booleans
        # if unsigned int values are preferred(0 for False, 1 for True), replace the 'dtype=np.bool' argument with 'dtype=np.uint' as shown
        # the commented line below illustrates this for digital input data; the same can be done for digital out

        # data['board_dig_in_data'] = np.zeros([header['num_board_dig_in_channels'], num_board_dig_in_samples], dtype=np.uint)
        data['board_dig_in_data'] = np.zeros([header['num_board_dig_in_channels'], num_board_dig_in_samples],
                                             dtype=np.bool)
        data['board_dig_in_raw'] = np.zeros(num_board_dig_in_samples, dtype=np.uint)

        data['board_dig_out_data'] = np.zeros([header['num_board_dig_out_channels'], num_board_dig_out_samples],
                                              dtype=np.bool)
        data['board_dig_out_raw'] = np.zeros(num_board_dig_out_samples, dtype=np.uint)

        # Read sampled data from file.
        print('Reading data from file...')

        # Initialize indices used in looping
        indices = {}
        indices['amplifier'] = 0
        indices['aux_input'] = 0
        indices['supply_voltage'] = 0
        indices['board_adc'] = 0
        indices['board_dig_in'] = 0
        indices['board_dig_out'] = 0

        print_increment = 10
        percent_done = print_increment
        for i in range(num_data_blocks):
            read_one_data_block(data, header, indices, fid)

            # Increment indices
            indices['amplifier'] += header['num_samples_per_data_block']
            indices['aux_input'] += int(header['num_samples_per_data_block'] / 4)
            indices['supply_voltage'] += 1
            indices['board_adc'] += header['num_samples_per_data_block']
            indices['board_dig_in'] += header['num_samples_per_data_block']
            indices['board_dig_out'] += header['num_samples_per_data_block']

            fraction_done = 100 * (1.0 * i / num_data_blocks)
            if fraction_done >= percent_done:
                print('{}% done...'.format(percent_done))
                percent_done = percent_done + print_increment

        # Make sure we have read exactly the right amount of data.
        bytes_remaining = filesize - fid.tell()
        if bytes_remaining != 0: raise Exception('Error: End of file not reached.')

    # Close data file.
    fid.close()

    if (data_present):
        print('Parsing data...')

        # Extract digital input channels to separate variables.
        for i in range(header['num_board_dig_in_channels']):
            data['board_dig_in_data'][i, :] = np.not_equal(
                np.bitwise_and(data['board_dig_in_raw'], (1 << header['board_dig_in_channels'][i]['native_order'])), 0)

        # Extract digital output channels to separate variables.
        for i in range(header['num_board_dig_out_channels']):
            data['board_dig_out_data'][i, :] = np.not_equal(
                np.bitwise_and(data['board_dig_out_raw'], (1 << header['board_dig_out_channels'][i]['native_order'])),
                0)

        # Scale voltage levels appropriately.
        data['amplifier_data'] = np.multiply(0.195,
                                             (data['amplifier_data'].astype(np.int32) - 32768))  # units = microvolts
        data['aux_input_data'] = np.multiply(37.4e-6, data['aux_input_data'])  # units = volts
        data['supply_voltage_data'] = np.multiply(74.8e-6, data['supply_voltage_data'])  # units = volts
        if header['eval_board_mode'] == 1:
            data['board_adc_data'] = np.multiply(152.59e-6,
                                                 (data['board_adc_data'].astype(np.int32) - 32768))  # units = volts
        elif header['eval_board_mode'] == 13:
            data['board_adc_data'] = np.multiply(312.5e-6,
                                                 (data['board_adc_data'].astype(np.int32) - 32768))  # units = volts
        else:
            data['board_adc_data'] = np.multiply(50.354e-6, data['board_adc_data'])  # units = volts
        data['temp_sensor_data'] = np.multiply(0.01, data['temp_sensor_data'])  # units = deg C

        # Check for gaps in timestamps.
        num_gaps = np.sum(np.not_equal(data['t_amplifier'][1:] - data['t_amplifier'][:-1], 1))
        if num_gaps == 0:
            print('No missing timestamps in data.')
        else:
            print('Warning: {0} gaps in timestamp data found.  Time scale will not be uniform!'.format(num_gaps))

        # Scale time steps (units = seconds).
        data['t_amplifier'] = data['t_amplifier'] / header['sample_rate']
        data['t_aux_input'] = data['t_amplifier'][range(0, len(data['t_amplifier']), 4)]
        data['t_supply_voltage'] = data['t_amplifier'][
            range(0, len(data['t_amplifier']), header['num_samples_per_data_block'])]
        data['t_board_adc'] = data['t_amplifier']
        data['t_dig'] = data['t_amplifier']
        data['t_temp_sensor'] = data['t_supply_voltage']

        # If the software notch filter was selected during the recording, apply the
        # same notch filter to amplifier data here.
        if header['notch_filter_frequency'] > 0 and header['version']['major'] < 3:
            print('Applying notch filter...')

            print_increment = 10
            percent_done = print_increment
            for i in range(header['num_amplifier_channels']):
                data['amplifier_data'][i, :] = notch_filter(data['amplifier_data'][i, :], header['sample_rate'],
                                                            header['notch_filter_frequency'], 10)

                fraction_done = 100 * (i / header['num_amplifier_channels'])
                if fraction_done >= percent_done:
                    print('{}% done...'.format(percent_done))
                    percent_done += print_increment
    else:
        data = [];

    # Move variables to result struct.
    result = data_to_result(header, data, data_present)

    print('Done!  Elapsed time: {0:0.1f} seconds'.format(time.time() - tic))
    return result



def plural(n):
    """Utility function to optionally pluralize words based on the value of n.
    """

    if n == 1:
        return ''
    else:
        return 's'



def sorting_day_split(sorting, date_id_all, day_length, pack_folder,
                      sorting_save_name='firings_inlier'):
    sampling_freq = sorting.get_sampling_frequency()

    fig, ax = plt.subplots(1, 1, figsize=(20, 40))
    sw.plot_rasters(sorting, time_range=(0, np.sum(day_length)), ax=ax)

    colors = []
    cm = pylab.get_cmap('rainbow')
    NUM_COLORS = len(day_length)
    for i in range(NUM_COLORS):
        colors.append(cm(1. * i / NUM_COLORS))  # color will now be an RGBA tuple

    for i in range(len(day_length)):
        ax.axvspan(np.sum(day_length[:i]) / sampling_freq, np.sum(day_length[:(i + 1)]) / sampling_freq,
                   facecolor=colors[i], alpha=0.1)
    plt.show()
    plt.savefig(pack_folder + '/sorting/' + sorting_save_name + '.png', dpi=300)

    for i in range(len(day_length)):
        pack_folder_i = pack_folder + '/' + date_id_all[i] + '/'

        if os.path.exists(pack_folder_i) == False:
            os.mkdir(pack_folder_i)

        start_frame = np.sum(day_length[:i])
        end_frame = np.sum(day_length[:(i + 1)])

        sub_sorting = sorting.frame_slice(start_frame, end_frame)

        keep_unit_ids = []
        for unit_id in sub_sorting.unit_ids:
            spike_train = sub_sorting.get_unit_spike_train(unit_id=unit_id)
            n = spike_train.size
            if (n > 20):
                keep_unit_ids.append(unit_id)

        curated_sub_sorting = sub_sorting.select_units(unit_ids=keep_unit_ids, renamed_unit_ids=None)

        save_path = pack_folder_i + '/sorting/' + sorting_save_name + '.npz'
        if os.path.exists(pack_folder_i + '/sorting/') == False:
            os.mkdir(pack_folder_i + '/sorting/')
        NpzSortingExtractor.write_sorting(curated_sub_sorting, save_path)



def create_mesh_probe(n):
    positions = np.zeros((n, 2))
    for i in range(n):
        x = i // 8
        y = i % 8
        positions[i] = x, y
    positions *= 300

    test = [positions[i] for i in range(8, 16)]
    test1 = copy.deepcopy(test[::-1])
    for idx, i in enumerate(range(8, 16)):
        positions[i] = test1[idx]

    test = [positions[i] for i in range(24, 32)]
    test1 = copy.deepcopy(test[::-1])
    for idx, i in enumerate(range(24, 32)):
        positions[i] = test1[idx]

    mesh_probe = Probe(ndim=2, si_units='um')
    mesh_probe.set_contacts(positions=positions, shapes='circle', shape_params={'radius': 5})

    ant = {'first_index': 0}
    mesh_probe.annotate(**ant)
    channel_indices_raw = np.arange(32)
    channel_indices = [i for i in channel_indices_raw]
    mesh_probe.set_device_channel_indices(channel_indices)
    return mesh_probe



def stack_recordings(pack_folder_pre, mesh_probe, trigger_val=2):

    cont_data_all = []
    cont_trigger_all = []
    for dirpath, dirname, filenames in os.walk(pack_folder_pre):
        for i in filenames:

            if '.rhd' in i:
                print(dirpath+'/'+i)

                raw_data=read_data(dirpath+'/'+i)

                sampling_freq = raw_data['frequency_parameters']['amplifier_sample_rate']
                test=raw_data['board_adc_data']
#                 test[test<trigger_val]=0
#                 test[test>=trigger_val]=1
                cont_trigger_all.append(test)
                cont_data_all.append(raw_data['amplifier_data'].T)

    cont_data_all = np.vstack(cont_data_all)
    cont_trigger_all = np.concatenate(cont_trigger_all,axis=1)

    np_mean = np.mean(cont_data_all)
    np_std = np.std(cont_data_all)
    threshold_neg = -300
    threshold_pos = 300

    neg_indices = np.where(cont_data_all<threshold_neg)
    pos_indices = np.where(cont_data_all>threshold_pos)


    remove_indices_neg_0 = [neg_indices[0]+i for i in np.arange(-10,10)]
    remove_indices_neg_1 = [neg_indices[1] for i in np.arange(-10,10)]

    remove_indices_pos_0 = [pos_indices[0]+i for i in np.arange(-10,10)]
    remove_indices_pos_1 = [pos_indices[1] for i in np.arange(-10,10)]

    remove_indices_0 = np.array(np.append(remove_indices_neg_0,remove_indices_pos_0)).flatten()
    remove_indices_1 = np.array(np.append(remove_indices_neg_1,remove_indices_pos_1)).flatten()


    remove_indices_0_temp = np.where(remove_indices_0>0, remove_indices_0, 0)
    remove_indices_0 = np.where(remove_indices_0_temp<cont_data_all.shape[0]-1, remove_indices_0_temp, cont_data_all.shape[0]-1)

    cont_data_all[remove_indices_0,remove_indices_1] = np_mean


    recording = se.NumpyRecording(traces_list=cont_data_all,
                              sampling_frequency=sampling_freq)

    recording.set_probe(mesh_probe, in_place=True)

    return recording, cont_data_all,cont_trigger_all


def read_intan_data_from_dir(date_id_all, pack_folder_ix=['./', '22_C6']):
    recording_traces = []
    session_length_concat = []
    day_length = []
    cont_trigger_all_all = []
    save_folder_name = '_'.join(date_id_all)
    data_folder_all = f'./processed_data/Ephys_concat_{save_folder_name}/'

    if os.path.exists(data_folder_all):
        recording_concat = spikeinterface.core.base.BaseExtractor.load_from_folder(data_folder_all)
        session_length_concat = np.load(data_folder_all + 'session_length.npy')
        day_length = np.load(data_folder_all + 'day_length.npy')
        cont_trigger_all_all = np.load(data_folder_all + 'cont_trigger_all_all.npy')
    else:
        for date_id in date_id_all:
            pack_folder_pre = pack_folder_ix[0] + date_id + pack_folder_ix[1]

            data_folder_pre = f'./processed_data/Ephys_{date_id}/'

            if os.path.exists(data_folder_pre):
                recording = spikeinterface.core.base.BaseExtractor.load_from_folder(data_folder_pre)
                session_length = np.load(data_folder_pre + 'session_length.npy')
                cont_trigger_all = np.load(data_folder_pre + 'cont_trigger_all.npy')
            else:
                mesh_probe = create_mesh_probe(32)
                # plot_probe(mesh_probe,with_channel_index=True)

                recording, session_length, cont_trigger_all = stack_recordings(pack_folder_pre, mesh_probe,
                                                                               trigger_val=3)
                #             recording.set_probe(mesh_probe, in_place=True)

                recording.set_probe(mesh_probe, in_place=True)

                recording = recording.save(folder=data_folder_pre)
                np.save(data_folder_pre + 'session_length.npy', session_length)
                np.save(data_folder_pre + 'cont_trigger_all.npy', cont_trigger_all)

            sampling_freq = recording.get_sampling_frequency()
            recording_trace = recording.get_traces()
            recording_traces.append(recording_trace)
            session_length_concat.append(session_length)
            cont_trigger_all_all.append(cont_trigger_all)
            print(recording_trace.shape)
            day_length.append(recording_trace.shape[0])

        recording_traces = np.vstack(recording_traces)
        session_length_concat = np.vstack(session_length_concat)

        recording_concat = se.NumpyRecording(traces_list=recording_traces,
                                             sampling_frequency=sampling_freq)
        recording_concat.set_probe(mesh_probe, in_place=True)

        recording_concat = recording_concat.save(folder=data_folder_all)

        np.save(data_folder_all + 'session_length.npy', session_length_concat)
        np.save(data_folder_all + 'day_length.npy', day_length)
        np.save(data_folder_all + 'cont_trigger_all_all.npy', cont_trigger_all_all)

    print(recording_concat)
    print('Num. channels = {}'.format(len(recording_concat.get_channel_ids())))
    print('Sampling frequency = {} Hz'.format(recording_concat.get_sampling_frequency()))
    print('Num. timepoints seg0= {}'.format(recording_concat.get_num_segments()))

    return recording_concat, session_length_concat, day_length, cont_trigger_all_all




def read_ns4datafile(datafile):
    brpylib_ver_req = "1.3.1"
    if brpylib_ver.split('.') < brpylib_ver_req.split('.'):
        raise Exception("requires brpylib " + brpylib_ver_req + " or higher, please use latest version")
    elec_ids = 'all'  # 'all' is default for all (1-indexed)                                        
    nsx_file = NsxFile(datafile)
    # Extract data - note: data will be returned based on *SORTED* elec_ids, see cont_data['elec_ids']
    cont_data = nsx_file.getdata(elec_ids)
    unit = nsx_file.extended_headers[1]['Units']
    # Close the nsx file now that all data is out
    nsx_file.close()
    return cont_data, unit



def detect_spike(trace0_car,thr_min = 5, distance=30, ch_max_simul_firing = 3):
#     noise_std_detect = np.median(abs(trace0_car - np.mean(trace0_car, axis=0)[None, :]) / 0.6745, axis=0)
    noise_std_detect = np.median(abs(trace0_car)/0.6745, axis=0)

    thr = thr_min * noise_std_detect
    thrmax = 30 * noise_std_detect

    spikes = np.zeros(trace0_car.shape)
    for i in range(noise_std_detect.shape[0]):
        peaks, _ = scipy.signal.find_peaks(-trace0_car[:, i], thr[i], distance=distance)
        spikes[peaks, i] = 1

    print(np.sum(spikes == 1))
    # larger value no more than thrmax
    points = trace0_car.shape[0]
    spike_coord = np.argwhere(spikes == 1)
    for i in range(spike_coord.shape[0]):
        near_start = spike_coord[i, 0] - 5
        near_end = spike_coord[i, 0] + 5
        if near_start < 0:
            near_start = 0
        if near_end >= points:
            near_end = points - 1
        if np.any(np.max(trace0_car[near_start:near_end, :], axis=0) >= thrmax):
            spikes[spike_coord[i, 0], spike_coord[i, 1]] = 0

    # no simultanous firing!!!!
    thres_cross = ch_max_simul_firing
    spikes[np.sum(spikes, axis=1) > thres_cross, :] = 0
    print(np.sum(spikes == 1))

    return spikes



def list_substraction(a, b):

    ca = Counter(a)
    cb = Counter(b)

    result_b = sorted((cb - ca).elements())
    return result_b





def offline_smoother(yIn, kernSD, stepSize):
    causal = False;

    if (kernSD == 0) or (yIn.shape[1]==1):
        yOut = yIn
        return yOut


    # Filter half length
    # Go 3 standard deviations out
    fltHL = math.ceil(3 * kernSD / stepSize)

    # Length of flt is 2*fltHL + 1
    flt = scipy.stats.norm.pdf(np.arange(-fltHL*stepSize ,fltHL*stepSize,stepSize), loc=0, scale = kernSD)


    [yDim, T] = shape(yIn)
    yOut      = np.zeros((yDim, T))

    nm = scipy.signal.convolve(flt.reshape(1,-1), np.ones((1,T)), mode='full', method='auto')

    for i in range(yDim):

        ys = np.divide(scipy.signal.convolve(flt.reshape(1,-1),yIn[i,:].reshape(1,-1)) , nm)
    #     # Cut off edges so that result of convolution is same length
    #     # as original data
        yOut[i,:] = ys[0,fltHL:ys.shape[1]-fltHL+1]
    return yOut




class FSM2:
    def __init__(self, k, d, tau0=None, Minv0=None, W0=None, learning_rate=None):
        self.k = k
        self.d = d
        if W0 is None:
            self.W = np.eye(k, d) / d
        else:
            self.W = W0;
        self.M = np.eye(k)
        if Minv0 is None:
            self.Minv = np.eye(k)
        else:
            self.Minv = Minv0
        if tau0 is None:
            self.tau = 0.5
        else:
            self.tau = tau0
        if learning_rate is None:
            self.lr = lambda x: 1.0 / (2 * (x + 100) + 5)
        else:
            self.lr = lambda x: learning_rate

        self.t = 0
        self.outer_W = np.zeros(shape(self.W))
        self.outer_Minv = np.zeros(shape(self.Minv))
        self.y = []

    def fit_next(self, x):
        step = self.lr(self.t)
        #         print('step',step)

        z = np.dot(self.W, x.T)
        y = np.zeros(shape(z))
        #         print('z',z)
        #         print('y',y)

        # The loop below solves the linear system My = Wx
        y[0] = z[0] / self.M[0, 0]
        #         print('y[0]',y[0])
        for i in np.arange(1, self.k):
            y[i] = z[i]
            #             print('i',i,'y[i]',y[i])
            for j in np.arange(0, i):
                y[i] = y[i] - y[j] * self.M[i, j]
            #                 print('i',i,'j',j,'y[i]',y[i])

            y[i] = y[i] / self.M[i, i]
        #             print('i',i,'y[i]',y[i])

        y_up = np.tril(
            np.dot(y.reshape(y.shape[0], 1), y.reshape(y.shape[0], 1).T))  # extract lower triangular component of y
        #         print('y_up',y_up)
        self.W = self.W + step * (y.reshape(y.shape[0], 1) * x.reshape(x.shape[0], 1).T - self.W)  # update W matrix
        #         print('W',self.W)
        self.M = self.M + step * (y_up - self.M)  # update M matrix
        #         print('M',self.M)
        self.t = self.t + 1;
        #         print('t',self.t)
        return y






#! /bin/env python
#
# Michael Gibson 23 April 2015
# Modified Adrian Foy Sep 2018

import sys, struct
from intanutil.qstring import read_qstring

def read_header(fid):
    """Reads the Intan File Format header from the given file."""

    # Check 'magic number' at beginning of file to make sure this is an Intan
    # Technologies RHD2000 data file.
    magic_number, = struct.unpack('<I', fid.read(4)) 
    if magic_number != int('c6912702', 16): raise Exception('Unrecognized file type.')

    header = {}
    # Read version number.
    version = {}
    (version['major'], version['minor']) = struct.unpack('<hh', fid.read(4)) 
    header['version'] = version

    print('')
    print('Reading Intan Technologies RHD2000 Data File, Version {}.{}'.format(version['major'], version['minor']))
    print('')

    freq = {}

    # Read information of sampling rate and amplifier frequency settings.
    header['sample_rate'], = struct.unpack('<f', fid.read(4))
    (freq['dsp_enabled'], freq['actual_dsp_cutoff_frequency'], freq['actual_lower_bandwidth'], freq['actual_upper_bandwidth'], 
    freq['desired_dsp_cutoff_frequency'], freq['desired_lower_bandwidth'], freq['desired_upper_bandwidth']) = struct.unpack('<hffffff', fid.read(26))


    # This tells us if a software 50/60 Hz notch filter was enabled during
    # the data acquisition.
    notch_filter_mode, = struct.unpack('<h', fid.read(2))
    header['notch_filter_frequency'] = 0
    if notch_filter_mode == 1:
        header['notch_filter_frequency'] = 50
    elif notch_filter_mode == 2:
        header['notch_filter_frequency'] = 60
    freq['notch_filter_frequency'] = header['notch_filter_frequency']

    (freq['desired_impedance_test_frequency'], freq['actual_impedance_test_frequency']) = struct.unpack('<ff', fid.read(8))

    note1 = read_qstring(fid)
    note2 = read_qstring(fid)
    note3 = read_qstring(fid)
    header['notes'] = { 'note1' : note1, 'note2' : note2, 'note3' : note3}

    # If data file is from GUI v1.1 or later, see if temperature sensor data was saved.
    header['num_temp_sensor_channels'] = 0
    if (version['major'] == 1 and version['minor'] >= 1) or (version['major'] > 1) :
        header['num_temp_sensor_channels'], = struct.unpack('<h', fid.read(2))
        
    # If data file is from GUI v1.3 or later, load eval board mode.
    header['eval_board_mode'] = 0
    if ((version['major'] == 1) and (version['minor'] >= 3)) or (version['major'] > 1) :
        header['eval_board_mode'], = struct.unpack('<h', fid.read(2))
        
        
    header['num_samples_per_data_block'] = 60
    # If data file is from v2.0 or later (Intan Recording Controller), load name of digital reference channel
    if (version['major'] > 1):
        header['reference_channel'] = read_qstring(fid)
        header['num_samples_per_data_block'] = 128

    # Place frequency-related information in data structure. (Note: much of this structure is set above)
    freq['amplifier_sample_rate'] = header['sample_rate']
    freq['aux_input_sample_rate'] = header['sample_rate'] / 4
    freq['supply_voltage_sample_rate'] = header['sample_rate'] / header['num_samples_per_data_block']
    freq['board_adc_sample_rate'] = header['sample_rate']
    freq['board_dig_in_sample_rate'] = header['sample_rate']

    header['frequency_parameters'] = freq

    # Create structure arrays for each type of data channel.
    header['spike_triggers'] = []
    header['amplifier_channels'] = []
    header['aux_input_channels'] = []
    header['supply_voltage_channels'] = []
    header['board_adc_channels'] = []
    header['board_dig_in_channels'] = []
    header['board_dig_out_channels'] = []

    # Read signal summary from data file header.

    number_of_signal_groups, = struct.unpack('<h', fid.read(2))
    print('n signal groups {}'.format(number_of_signal_groups))

    for signal_group in range(1, number_of_signal_groups + 1):
        signal_group_name = read_qstring(fid)
        signal_group_prefix = read_qstring(fid)
        (signal_group_enabled, signal_group_num_channels, signal_group_num_amp_channels) = struct.unpack('<hhh', fid.read(6))

        if (signal_group_num_channels > 0) and (signal_group_enabled > 0):
            for signal_channel in range(0, signal_group_num_channels):
                new_channel = {'port_name' : signal_group_name, 'port_prefix' : signal_group_prefix, 'port_number' : signal_group}
                new_channel['native_channel_name'] = read_qstring(fid)
                new_channel['custom_channel_name'] = read_qstring(fid)
                (new_channel['native_order'], new_channel['custom_order'], signal_type, channel_enabled, new_channel['chip_channel'], new_channel['board_stream']) = struct.unpack('<hhhhhh', fid.read(12))
                new_trigger_channel = {}
                (new_trigger_channel['voltage_trigger_mode'], new_trigger_channel['voltage_threshold'], new_trigger_channel['digital_trigger_channel'], new_trigger_channel['digital_edge_polarity'])  = struct.unpack('<hhhh', fid.read(8))
                (new_channel['electrode_impedance_magnitude'], new_channel['electrode_impedance_phase']) = struct.unpack('<ff', fid.read(8))

                if channel_enabled:
                    if signal_type == 0:
                        header['amplifier_channels'].append(new_channel)
                        header['spike_triggers'].append(new_trigger_channel)
                    elif signal_type == 1:
                        header['aux_input_channels'].append(new_channel)
                    elif signal_type == 2:
                        header['supply_voltage_channels'].append(new_channel)
                    elif signal_type == 3:
                        header['board_adc_channels'].append(new_channel)
                    elif signal_type == 4:
                        header['board_dig_in_channels'].append(new_channel)
                    elif signal_type == 5:
                        header['board_dig_out_channels'].append(new_channel)
                    else:
                        raise Exception('Unknown channel type.')
                        
    # Summarize contents of data file.
    header['num_amplifier_channels'] = len(header['amplifier_channels'])
    header['num_aux_input_channels'] = len(header['aux_input_channels'])
    header['num_supply_voltage_channels'] = len(header['supply_voltage_channels'])
    header['num_board_adc_channels'] = len(header['board_adc_channels'])
    header['num_board_dig_in_channels'] = len(header['board_dig_in_channels'])
    header['num_board_dig_out_channels'] = len(header['board_dig_out_channels'])

    return header

if __name__ == '__main__':
    h=read_header(open(sys.argv[1], 'rb'))
    print(h)



from setuptools import setup, find_packages

with open('README.md','r') as f:
    description = f.read()

setup(
    name='autosort_neuron',
    version='0.0.1.1',
    packages=find_packages(),
    install_requires=[
        'spikeinterface==0.95.1',
        'torch==1.12.1',
        'pandas'
    ],
    
    long_description=description,
    long_description_content_type='text/markdown',

)